import math
import time
from collections.abc import Callable

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import (
    AsyncConnection,
    AsyncEngine,
    create_async_engine,
)

from grammdb.contracts import (
    DatabaseModule,
    SchemaModule,
)

_timestamp = lambda: math.floor(time.time())


async def init_db(
    db_mod: DatabaseModule,
    db_uri: str,
    *,
    drop_tables: bool = False,
    hook: Callable[[AsyncEngine], None] = lambda x: None,
    **engine_options,
) -> None:
    """Setup the sqlalchemy AsyncEngine for the provided database module to connect to the server at
    `db_uri`.


    Args:
        db_mod: The database module to setup.
            This stateful module will hold the newly configured engine.
        db_uri: The connection string for the target backend.
        drop_tables: Keyword-only argument.
            If True, drop and recreate all tables in the `db_mod.schema`.
            If False, only create tables that don't already exist.
            Default: False
        hook: Keyword-only argument.
            Callback function to invoke after creating the engine, but before creating tables
            and assigning it to the database module. Default is a no-op.
            Accepts a single argument, which is the newly created engine.
            This can be used to apply custom setup code on the engine. For example, to enable
            sqlite foreign key support:
            ```pycon
            >>> import sqlalchemy as sa
            >>> def setup_hook(engine: sa.ext.asyncio.AsyncEngine) -> None:
            ...     def _(async_conn, dbapi_conn, conn = None):
            ...         async_conn.cursor().execute("PRAGMA foreign_keys = ON;")
            ...     sa.event.listen(engine.sync_engine, "connect", _)
            >>> async def ex():
            ...     await init_db(database(), "sqlite+aiosqlite://", hook=setup_hook)
            >>> asyncio.run(ex())

            ```
        **engine_options:
            Any additional keyword arguments are passed to `create_async_engine` as-is.
            Default:

                echo: False
                poolclass: `sqlalchemy.pool.AsyncAdaptedQueuePool`
                pool_size: 10

            Example:
                ```pycon
                >>> async def ex():
                ...     await init_db(database(), "sqlite+aiosqlite://", pool_size=1)
                ...     return database().get_engine().pool.size()
                >>> asyncio.run(ex())
                1

                ```

    """

    default_options = {"echo": False, "poolclass": sa.pool.AsyncAdaptedQueuePool, "pool_size": 10}

    engine = create_async_engine(db_uri, **{**default_options, **engine_options})

    for table in db_mod.get_schema().get_metadata().tables.values():
        table.append_column(
            sa.Column("created_at", sa.Integer, default=_timestamp), replace_existing=True
        )
        table.append_column(
            sa.Column("updated_at", sa.Integer, nullable=True), replace_existing=True
        )

    hook(engine)
    db_mod.set_engine(engine)
    async with engine.begin() as conn:
        if drop_tables:
            await conn.run_sync(db_mod.get_schema().get_metadata().drop_all)
        await conn.run_sync(db_mod.get_schema().get_metadata().create_all)


class _DatabaseFacade:
    """Concrete implementation of the `grammdb.contracts.DatabaseModule` Protocol."""

    _ENGINE: AsyncEngine | None = None
    _SCHEMA: SchemaModule

    def __init__(self, schema_mod: SchemaModule) -> None:
        """
        Args:
            schema_mod: The `grammdb.contracts.DatabaseModule.schema`.
        """

        self._SCHEMA = schema_mod

    def get_schema(self) -> SchemaModule:
        return self._SCHEMA

    def get_engine(self) -> AsyncEngine:
        assert (
            self._ENGINE is not None
        ), "No engine found. Did you forget to call initialize on this database?"
        return self._ENGINE

    def set_engine(self, value: AsyncEngine) -> None:
        self._ENGINE = value

    def new_connection(self) -> AsyncConnection:
        return self.get_engine().connect()


def db_factory(schema_mod: SchemaModule) -> Callable[[], DatabaseModule]:
    """Returns a closure that provides a valid `grammdb.contracts.DatabaseModule`.

    The closure caches the database module, so subsequent calls to the returned function
    will return the same object.

    In other words, the following invariants hold:
        ```pycon
        >>> db1 = db_factory(schema)
        >>> db2 = db_factory(schema)
        >>> db1() is db1()
        True
        >>> db2() is db2()
        True
        >>> db1() is not db2()
        True

        ```

    Note:
        The concrete type of the object returned by the closure can be any valid implementation
        of the `grammdb.contracts.DatabaseModule` protocol, including arbitrary classes or modules.
        The only guarantee about this object is that it implements that protocol correctly, so the
        concrete type and any fields on that object that are not part of the DatabaseModule protocol
        should be treated as implementation details that could change at any time. This is true even
        for public members of the concrete type that are not part of the DatabaseModule interface.
        Red flags for client code are `isinstance` checks and any usage of a field not specified
        in the protocol.
    """

    db = _DatabaseFacade(schema_mod)

    def _() -> DatabaseModule:
        return db

    return _



__all__ = [
    "close_transaction",
    "commit_transaction",
    "connection_ctx",
    "contracts",
    "db_factory",
    "delete",
    "exceptions",
    "init_db",
    "insert",
    "rollback_transaction",
    "select",
    "start_transaction",
    "types",
    "update",
]

from . import (
    contracts,
    exceptions,
    types,
)
from .database import (
    db_factory,
    init_db,
)
from .operations import (
    delete,
    insert,
    select,
    update,
)
from .transactions import (
    close_transaction,
    commit_transaction,
    connection_ctx,
    rollback_transaction,
    start_transaction,
)

__version__ = "0.0.1"



__all__ = [
    "delete",
    "insert",
    "select",
    "update",
]

import functools
import math
import time
from typing import Any

import sqlalchemy as sa

from grammdb.types import WhereFunc

_reducer = lambda stmt, where_func: where_func(stmt)
_timestamp = lambda: math.floor(time.time())


def select(
    columns: str | tuple[str, ...], from_clause: sa.Select, *where_clauses: WhereFunc
) -> sa.Select:
    """Wrapper for generating sqlalchemy select statements with a more sql-like DSL.

    Note:
        Using "*" as the first argument is a special case that does not actually do a select star
        in SQL. It is simply syntax sugar around SQLAlchemy's regular behavior of selecting the
        full set of columns in the from clause by name.
        In other words, `sa.select(MyTable)` and `grammdb.select("*", sa.select(MyTable))` are
        effectively equivalent.

    Args:
        columns:
            One of the following:
                A single string representing the name of the column to select.
                A literal "*" meaning select all columns in the from clause.
                A list of strings representing the column names to select from the from clause.
        from_clause:
            A selectable object. Usually, this would be provided using a from function to create
            a statement that reads more like sql.
        *where_clauses:
            Arbitrary number of where functions to apply as the where clause in the built query.
    """

    # Explanation for Any:
    # SQLAlchemy's core is too dynamic, and it's type annotations don't match the documented
    # inputs. E.g. a tuple of string column names doesn't satisfy the type annotations for
    # with_only_columns below, but this works at runtime.
    select_columns: Any = (
        from_clause.selected_columns if columns == "*" else from_clause.selected_columns[columns]
    )
    return functools.reduce(
        _reducer,
        where_clauses,
        from_clause.with_only_columns(select_columns),
    )


def update(table_: sa.Table, *where_clauses: WhereFunc, **set_clause: Any) -> sa.Update:
    """Wrapper for generating sqlalchemy update statements with a more sql-like DSL."""

    return functools.reduce(
        _reducer,
        where_clauses,
        sa.update(table_),
    ).values(**set_clause, updated_at=_timestamp())


def delete(from_: sa.Table, *where_clauses: WhereFunc) -> sa.Delete:
    """Wrapper for generating sqlalchemy delete statements with a more sql-like DSL."""

    return functools.reduce(
        _reducer,
        where_clauses,
        sa.delete(from_),
    )


def insert(*, into: sa.Table, **values: Any) -> sa.Insert:
    """Wrapper for generating sqlalchemy insert statements with a more sql-like DSL.

    The `into` argument is enforced as keyword-only to make the statements read more like sql.
    """

    return sa.insert(into).values(**values)



from typing import Protocol

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import (
    AsyncConnection,
    AsyncEngine,
)


class SchemaModule(Protocol):
    """Protocol defining a database schema.

    This protocol is intended to be implemented as a python module and not
    a class, but a class that satisfies the protocol will work as well.
    """

    def get_metadata(self) -> sa.MetaData:
        """The sqlalchemy metadata object that contains the tables in this schema."""

        pass


class DatabaseModule(Protocol):
    """Protocol defining a database wrapper.

    For convenience, a class or module satisfying this protocol is returned
    by the closure created by the `grammdb.db_factory` function, but
    any class or module that correctly implements this protocol can be used.
    """

    def get_schema(self) -> SchemaModule:
        """The schema defining the tables in this database."""

        pass

    def get_engine(self) -> AsyncEngine:
        """The underlying engine that gets set up by the
        `grammdb.init_db` function."""

        pass

    def set_engine(self, engine: AsyncEngine):
        """Assigns the underlying engine that will be returned by `get_engine`.

        This function is used by the `grammdb.init_db` function and generally
        should not be called by client code.
        """

        pass

    def new_connection(self) -> AsyncConnection:
        """Factory for creating new connections in the underlying engine's connection
        pool.

        This function is used by the transaction management helper functions, and
        is not intended to be called directly by client code.
        """

        pass



import contextlib

from sqlalchemy.exc import IntegrityError


class ConstraintError(Exception):
    """Base class for constraint violation exceptions."""

    pass


class UniqueConstraintError(ConstraintError):
    """Represents a unique constraint violation."""

    pass


class ForeignKeyConstraintError(ConstraintError):
    """Represents a foreign key constraint violation."""

    pass


class CheckConstraintError(ConstraintError):
    """Represents a check constraint violation."""

    pass


@contextlib.contextmanager
def constraint_error():
    """Context manager that will catch and parse a sqlalchemy.exc.IntegrityError
    and raise one of the specific constraint errors from this module in its place.

    Note:
        This is only tested on sqlite3 and postgresql. This function relies on the
        string message provided in the sqlalchemy error to determine which constraint
        was violated, and this may not be reliable on all backends.

    Example:
        ```pycon
        >>> from grammdb.exceptions import constraint_error, UniqueConstraintError
        >>> def generate_xkcd_random_number():
        ...     # Chosen by fair dice roll.
        ...     # Guaranteed to be random.
        ...     return 4
        >>> async def try_insert(rand_num):
        ...     try:
        ...         async with grammdb.connection_ctx(database()) as conn:
        ...             with constraint_error():
        ...                 await conn.execute(grammdb.insert(into=schema.MyTable, num=rand_num))
        ...                 await conn.commit()
        ...     except UniqueConstraintError:
        ...         print("Random number collision encountered. Trying a new random number...")
        ...         await try_insert(generate_xkcd_random_number())
        ...     except Exception as e:
        ...         print(f"Unknown error occurred: {e}")
        >>> async def ex():
        ...     await grammdb.init_db(database(), "sqlite+aiosqlite://")
        ...     await try_insert(generate_xkcd_random_number())
        >>> asyncio.run(ex())

        ```
    """

    try:
        yield None
    except IntegrityError as e:
        err_str = str(e)
        comp_str = err_str.lower()
        if "unique constraint" in comp_str:
            raise UniqueConstraintError(err_str) from e
        elif "check constraint" in comp_str:
            raise CheckConstraintError(err_str) from e
        elif "foreign key constraint" in comp_str:
            raise ForeignKeyConstraintError(err_str) from e
        else:
            raise e



from collections.abc import Callable

import sqlalchemy as sa

type WhereFunc[QUERY: (sa.Select, sa.Update, sa.Delete)] = Callable[[QUERY], QUERY]
"""Type alias used in the where clause arguments in the
`grammdb.select`, `grammdb.update`, and `grammdb.delete` functions.

The QUERY generic parameter is a tuple type (not a union) constrained to sqlalchemy's
Select, Update, or Delete constructs.

This means that this type alias represents a function that takes in a Select, Update, or Delete
statement and returns the same type of statement, and so is suitable for use in a reduction.
"""



from sqlalchemy.ext.asyncio import AsyncConnection

from grammdb.contracts import DatabaseModule


def connection_ctx(db: DatabaseModule) -> AsyncConnection:
    """Returns a connection object suitable for use as an async context manager.

    Example:
        ```pycon
        >>> import sqlalchemy as sa
        >>> async def ex():
        ...     await grammdb.init_db(database(), "sqlite+aiosqlite://")
        ...     async with connection_ctx(database()) as conn:
        ...         stmt = grammdb.insert(into=schema.MyTable, num=1).returning(sa.column("num"))
        ...         result = await conn.execute(stmt)
        ...         await conn.commit()
        ...         return result.scalar_one()
        >>> asyncio.run(ex())
        1

        ```
    """

    return db.new_connection()


async def start_transaction(db: DatabaseModule) -> AsyncConnection:
    """Return an open transaction that can be passed around without a context block.

    Must be closed manually using one of the `{rollback|close|commit}_transaction` functions.

    Example:
        ```pycon
        >>> import sqlalchemy as sa
        >>> from sqlalchemy.exc import IntegrityError
        >>> async def ex():
        ...     tr = await start_transaction(database())
        ...     try:
        ...         stmt = grammdb.insert(into=schema.MyTable, num=1).returning(sa.column("num"))
        ...         result = await tr.execute(stmt)
        ...     except IntegrityError as e:
        ...         await rollback_transaction(tr)
        ...         raise e
        ...     else:
        ...         await commit_transaction(tr)
        ...         return result.scalar_one()
        >>> asyncio.run(ex())
        1

        ```
    """

    conn = db.new_connection()
    await conn.start()
    await conn.begin()
    return conn


async def rollback_transaction(tr: AsyncConnection) -> None:
    """Rollback a transaction manually.

    Intended to be used on the connection returned by `start_transaction`.
    """

    await tr.rollback()
    await tr.close()


async def close_transaction(tr: AsyncConnection) -> None:
    """Close a transaction manually.

    Intended to be used on the connection returned by `start_transaction`.
    """

    await tr.close()


async def commit_transaction(tr: AsyncConnection) -> None:
    """Commit a transaction manually.

    Intended to be used on the connection returned by `start_transaction`.
    """

    await tr.commit()
    await tr.close()
